package org.jobimtext.api;



/*******************************************************************************
* Copyright 2012
* Copyright (c) 2012 IBM Corp.
* 
* and
* 
* FG Language Technologie
* Technische Universitaet Darmstadt
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*   http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
******************************************************************************/
/**
 * This interface is used to
 *         retrieve all contents that are needed for the distributional
 *         thesaurus and the contextual thesaurus. We therefore need to specify
 *         the value of a KEY and VALUE. Furthermore the similar words are
 *         specified with ORDER2
 * @author Martin Riedl (riedl@cs.tu-darmstadt.de) 
 * @param <TERM> The structure used for a term representation
 * @param <CONTEXTS> The structure used for a context representation 
 * @param <ORDER2LIST> The structure used for representation of distributional similarity lists
 * @param <ORDER1LIST> The structure used for representation of lists containing the significance scores and counts
 * @param <ISAS> The structure used for representation of lists of sense representations
 * @param <SENSES> The structure used for representation of lists of sense representations 
 * @param <SENSE_CUIS> The structure used for representation of lists of sense representations
 */
public interface IThesaurus<TERM, CONTEXTS, ORDER2LIST, ORDER1LIST, ISAS, SENSES,  SENSE_CUIS> {
	/**
	 * Connects to the data source
	 * @return true, if connection successful
	 */
	public boolean connect();

	/**
	 * Ends the connection to the data source
	 */
	public void destroy();
	
	/**
	 * Retrieve the similarity score between two terms. Note, that this score is not necessarily symmetric!
	 * @param t1 term1
	 * @param t2 term2
	 * @return the similarity score
	 */
	public Double getSimilarTermScore(TERM t1, TERM t2);
	
	/**
	 * Get a list of all similar terms for a given term.
	 * @param key the input term
	 * @return list of similar terms with scores
	 */
	public ORDER2LIST getSimilarTerms(TERM key);
	
	/**
	 * Get a list of similar terms for a given term, maximal number is specified
	 * @param key the input term
	 * @param numberOfEntries the maximal number of items to retrieve
	 * @return list of similar terms with scores
	 */
	public ORDER2LIST getSimilarTerms(TERM key, int numberOfEntries);
	
	/**
	 * Get a list of similar terms for a given term, minimal similarity score is specified
	 * @param key the input term
	 * @param threshold the minimal similarity score
	 * @return list of similar terms with scores
	 */
	public ORDER2LIST getSimilarTerms(TERM key, double threshold);
	
	/**
	 * Get a list of all similar contexts for a given context
	 * @param values the input context
	 * @return list of similar contexts with scores
	 */
	public ORDER2LIST getSimilarContexts(CONTEXTS values);
	
	/**
	 * Get a list of all similar contexts for a given context, maximal number is specified
	 * @param values the input context
	 * @param numberOfEntries the maximal number of items to retrieve
	 * @return list of similar contexts with scores
	 */
	public ORDER2LIST getSimilarContexts(CONTEXTS values, int numberOfEntries);
	
	/**
	 * Get a list of all similar contexts for a given context,  minimal similarity score is specified
	 * @param values the input context
	 * @param threshold the minimal similarity score
	 * @return list of similar contexts with scores
	 */
	public ORDER2LIST getSimilarContexts(CONTEXTS values, double threshold);
	
	
	/**
	 * Get the term count from the corpus
	 * @param key the input term
	 * @return the frequency count
	 */
	public Long getTermCount(TERM key);
	
	/**
	 * Get the context count from the corpus
	 * @param key the input context
	 * @return the frequency count
	 */
	public Long getContextsCount(CONTEXTS key);
	
	/**
	 * Get the term-context count from the corpus
	 * @param key the input term
	 * @param val the input context
	 * @return the frequency count
	 */
	public Long getTermContextsCount(TERM key, CONTEXTS val);
	
	/**
	 * Get the significance score of a term-context combination
	 * @param key the input term
	 * @param val the input context
	 * @return the significance score
	 */
	public Double getTermContextsScore(TERM key, CONTEXTS val);
	
	/**
	 * Get all term-context scores for a given term
	 * @param key the input term
	 * @return list of contexts, containing the significance scores and counts
	 */
	public ORDER1LIST getTermContextsScores(TERM key);
	
	/**
	 * Get the top n term-context scores for a given term
	 * @param key the input term
	 * @param numberOfEntries the maximum number of entries to return
	 * @return list of contexts, containing the significance scores and counts
	 */
	public ORDER1LIST getTermContextsScores(TERM key, int numberOfEntries);
	
	/**
	 * Get all term-context scores for a given term, using a score threshold
	 * @param key the input term
	 * @param threshold the minimal significance score
	 * @return list of contexts, containing the significance scores and counts
	 */
	public ORDER1LIST getTermContextsScores(TERM key, double threshold);

	/**
	 * Get all term-context scores for a given context
	 * @param key the input context
	 * @return list of terms, containing the significance scores and counts
	 */
	public ORDER1LIST getContextsTermScores(CONTEXTS key);
	
	/**
	 * Get the top n term-context scores for a given context
	 * @param key the input context
	 * @param numberOfEntries the maximum number of entries to return
	 * @return list of terms, containing the significance scores and counts
	 */
	public ORDER1LIST getContextsTermScores(CONTEXTS key, int numberOfEntries);
	
	/**
	 * Get all term-context scores for a given context, using a score threshold
	 * @param key the input context
	 * @param threshold the minimal significance score
	 * @return list of contexts, containing the significance scores and counts
	 */
	public ORDER1LIST getContextsTermScores(CONTEXTS key, double threshold);

	
	/**
	 * Get all senses for the input term, using the standard sense type
	 * @param key the input term
	 * @return a list of all senses for the input term
	 */
	public SENSES getSenses(TERM key) ;
	
	/**
	 * Get all senses for the input term, using the standard sense type
	 * @param key the input term
	 * @return a list of all senses for the input term
	 */
	public ISAS getIsas(TERM key) ;
	
	/**
	 * Get all senses for the input term, using the standard sense type
	 * @param key the input term
	 * @return a list of all senses for the input term
	 */
	public SENSE_CUIS getSenseCUIs(TERM key);
	
	/**
	 * Get all senses for the input term, with a specified sense type
	 * @param key the input term
	 * @param type the required sense type
	 * @return a list of all senses for the input term
	 */
	public SENSES getSenses(TERM key,String type) ;
	
	/**
	 * Get all senses for the input term, with a specified sense type
	 * @param key the input term
	 * @param type the required sense type
	 * @return a list of all senses for the input term
	 */
	public ISAS getIsas(TERM key,String type) ;
	
	/**
	 * Get all senses for the input term, with a specified sense type
	 * @param key the input term
	 * @param type the required sense type
	 * @return a list of all senses for the input term
	 */
	public SENSE_CUIS getSenseCUIs(TERM key,String type);
	
	/**
	 * Get all possible sense types available 
	 * @return a list of the sense types names
	 */
	public String[] getSensesTypes();
	
	/**
	 * Get the name of the standard sense inventory type
	 * @return the standard sense type
	 */
	public String getStandardSense();
	
	/**
	 * Checks whether the term is contained in the resource
	 * @param key input term
	 * @return true, if the term is contained
	 */
	public boolean isTermContained(TERM key);
	
	/**
	 * Checks whether a list of terms is contained in the resource
	 * @param key list of input terms that should be checked
	 * @return a list of the terms that are contained
	 */
	public TERM[] isTermContained(TERM... key);
	
	/**
	 * Checks whether the context is contained in the resource
	 * @param key the input context
	 * @return true, if the term is contained
	 */
	public boolean isContextsContained(CONTEXTS key);
	
	/**
	 * Checks which contexts from the input list are contained in the resource
	 * @param key list of contexts that are checked
	 * @return a list of the contexts that are contained
	 */
	public CONTEXTS[] isContextsContained(CONTEXTS... key);
	

}
