/*******************************************************************************
 * Copyright 2012
 * Copyright (c) 2012 IBM Corp.
 * 

 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.jobimtext.api.output;

import static org.apache.uima.fit.util.JCasUtil.select;
import static org.apache.uima.fit.util.JCasUtil.selectCovered;

import java.beans.PropertyDescriptor;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.zip.GZIPOutputStream;

import org.apache.pig.impl.util.Pair;
import org.apache.uima.UimaContext;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.component.JCasAnnotator_ImplBase;
import org.apache.uima.fit.descriptor.ConfigurationParameter;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;
import org.apache.uima.util.Logger;
import org.jobimtext.api.struct.DatabaseThesaurusDatastructure;
import org.jobimtext.api.struct.IThesaurusDatastructure;
import org.jobimtext.holing.type.Sentence;
import org.jobimtext.type.Sense;

/**
 * SenseOutput Writer for text files, based on the HolingOutputWriters. It
 * prints the document text and the identified senses in the text.
 * 
 * @author eruppert
 * 
 */
public class SenseOutputXML extends JCasAnnotator_ImplBase {

	public static final String PARAM_OUTPUT_PATH = "OutputPath";
	public static final String PARAM_INPUT_NAME = "InputName";
	public static final String PARAM_OUTPUT_SUFFIX = "OutputSuffix";
	public static final String PARAM_APPEND_OUTPUT = "AppendOutput";
	public static final String PARAM_DOCUMENT_ANNOTATION = "DocumentAnnotation";
	public static final String PARAM_DOCUMENT_ANNOTATION_ID = "DocumentAnnotationId";
	public static final String PARAM_SENTENCE_ANNOTATION = "SentenceAnnotation";
	
    // Distributional Thesaurus
    public static final String PARAM_DT_CONFIG="DtConfiguration";
    @ConfigurationParameter(name = PARAM_DT_CONFIG)
    private String dtConfig = null;

	@ConfigurationParameter(name = PARAM_SENTENCE_ANNOTATION)
	private Class<? extends Annotation> sentenceAnnotation = Sentence.class;

	@ConfigurationParameter(name = PARAM_APPEND_OUTPUT)
	private boolean appendOutput = false;

	protected String outputPath;
	protected String inputName;
	protected String outputSuffix;
	protected Logger logger;
	protected String documentAnnotationId;
	protected Class<Annotation> documentAnnotationClass;
	
	private IThesaurusDatastructure<String, String> dt;

	private boolean zip = false;

	private Map<String, WriterRecord> writerMap = new HashMap<String, WriterRecord>();

	// private int countThresh = 1073741824; // bytes = 1 GB
	// private int countThresh = 10485760; //bytes = 1 MB
	private int countThresh = 510485760; // bytes = ca 50 MB

	/**
	 * Initialize parameters, create output folders
	 */
	@Override
	public void initialize(UimaContext context)
			throws ResourceInitializationException {
		super.initialize(context);
		logger = context.getLogger();

		inputName = (String) context.getConfigParameterValue(PARAM_INPUT_NAME);
		outputPath = (String) context
				.getConfigParameterValue(PARAM_OUTPUT_PATH);
		outputSuffix = (String) context
				.getConfigParameterValue(PARAM_OUTPUT_SUFFIX);
		appendOutput = (boolean) context
				.getConfigParameterValue(PARAM_APPEND_OUTPUT);

		initAnnotationClass(context, PARAM_SENTENCE_ANNOTATION,
				sentenceAnnotation);

		String documentAnnotation = (String) context
				.getConfigParameterValue(PARAM_DOCUMENT_ANNOTATION);
		documentAnnotationId = (String) context
				.getConfigParameterValue(PARAM_DOCUMENT_ANNOTATION_ID);

		try {
			this.documentAnnotationClass = (Class<Annotation>) Class
					.forName(documentAnnotation);
		} catch (Exception e) {
			System.err.println("Document Annotation does not exists: "
					+ e.getMessage());
		}
		
		// initialize and connect to DT
				Object dtConfigParam=context.getConfigParameterValue(PARAM_DT_CONFIG);		
				if (dtConfigParam !=null){
					dtConfig = dtConfigParam.toString();
					dt = new DatabaseThesaurusDatastructure(dtConfig);
					dt.connect();
				} 

		// create folders
		File outputDir = new File(outputPath);
		if (!outputDir.exists()) {
			outputDir.mkdirs();
		}
	}

	/**
	 * Destructor for the SenseOutput writer. Closes all open Writers.
	 */
	@Override
	public void destroy() {
		super.destroy();
		for (WriterRecord rec : writerMap.values()) {
			closeWriter(rec.writer);
		}
		dt.destroy();
	}

	/**
	 * Close a Writer
	 * 
	 * @param w
	 *            Writer to close
	 */
	private void closeWriter(Writer w) {
		try {
			w.flush();
			w.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Close a Writer by filename
	 * 
	 * @param filename
	 *            the filename of the Writer that should be closed
	 */
	protected void closeWriter(String filename) {
		WriterRecord rec = writerMap.get(filename);
		closeWriter(rec.writer);
		writerMap.remove(filename);
	}

	private String getvalue(Annotation instance) {
		Object value;
		try {
			value = new PropertyDescriptor(documentAnnotationId,
					documentAnnotationClass).getReadMethod().invoke(instance);
			return value.toString();
		} catch (Exception e) {
			System.err.println("The attribute: " + documentAnnotationId
					+ " could not be found in object of class "
					+ documentAnnotationClass);
			e.printStackTrace();
		}
		return "_Attribute_not_found_";
	}

	@Override
	public void process(JCas aJCas) throws AnalysisEngineProcessException {
		synchronized (this) {
			if (aJCas.getDocumentText().compareTo("") == 0)
				return;
			logger.log(Level.INFO, "Start SenseOutputWriter");
			String unitId = "";
			if (inputName != null) {
				String inputID = new File(inputName).getName();
				if (inputID.contains(".")) {
					inputID = inputID.substring(0, inputID.indexOf('.'));
				}
				unitId = inputID;
			} else if (documentAnnotationClass != null) {
				Annotation doc = JCasUtil.selectSingle(aJCas,
						documentAnnotationClass);

				unitId = getvalue(doc);
				unitId = unitId.replace('\\', '-');
				unitId = unitId.replace('/', '-');
				unitId = unitId.replace('.', '-');
			} else {
				logger.log(Level.SEVERE,
						"Neither DocumentAnnotation parameter, nor InputName parameter specified");
				throw new AnalysisEngineProcessException();
			}

			if (!unitId.matches("[a-zA-Z0-9_-]*")) {
				logger.log(
						Level.WARNING,
						"unitId derived from input file name contains special characters! This might cause problems.");
			}

			HashSet<String> filesToClose = new HashSet<String>();

			String filename = unitId;
			if (outputSuffix != null) {
				filename += "_" + outputSuffix;
			}

			writeOutput(filename, aJCas.getDocumentText().replace("</mediawiki>", "").replace("</page>", ""));

			HashMap<String, HashMap<Integer, Integer>> senses = new HashMap<String, HashMap<Integer, Integer>>();
			for (Sense sense : select(aJCas, Sense.class)) {
				String word = sense.getCoveredText();
				int id = sense.getSenseId();
				HashMap<Integer, Integer> entry;
				if (senses.containsKey(word)) {
					entry = senses.get(word);
					if (entry.containsKey(id)) {
						entry.put(id, entry.get(id) + 1);
					} else {
						entry.put(id, 1);
					}
				} else {
					entry = new HashMap<Integer, Integer>();
					entry.put(id, 1);
				}
				senses.put(word, entry);

			}
			HashMap<String, Pair<Integer, Integer>> maxSenses = getMaxSenses(senses);
			
			writeOutput(filename, "<senses>", false);
			for (String s : maxSenses.keySet()) {
//				System.out.println(s+maxSenses.get(s).first);
				writeOutput(filename, s+maxSenses.get(s).first+" ", false);
			}
			writeOutput(filename, "</senses>");
			
			writeOutput(filename, "<hypernyms>", false);
			for (String s : maxSenses.keySet()) {
				List<String > isas = dt.getSenses(s+"#NN").get(maxSenses.get(s).first).getIsas();
				for (String i : isas) {
					writeOutput(filename, i.replace("_", " ").replace(":", " ")+", ", false);
				}
			}
			writeOutput(filename, "</hypernyms>");
			
			writeOutput(filename, "</page>\n</mediawiki>");
			filesToClose.add(filename);

			for (String fn : filesToClose) {
				closeWriter(fn);
			}
			logger.log(Level.INFO, "Finish SenseOutputWriter");
		}

	}

	/**
	 * 
	 * @param senses
	 */
	private HashMap<String, Pair<Integer,Integer>> getMaxSenses(HashMap<String, HashMap<Integer, Integer>> senses) {
		HashMap<String, Pair<Integer,Integer>> ret = new HashMap<String, Pair<Integer,Integer>>(); 
		for (String n : senses.keySet()) {
			HashMap<Integer, Integer> entry = senses.get(n);
			int max = 0;
			int maxInd = -1;
			for (int i : entry.keySet()) {
				if (i < 0)
					continue;
				if (entry.get(i) > max) {
					max = entry.get(i);
					maxInd = i;
				}
			}
			if (maxInd >= 0) {
				ret.put(n, new Pair<Integer,Integer>(maxInd,max));;
			}
		}
		return ret;
	}

	private void initFileForModel(String model, int part) throws IOException {
		String path = outputPath + "/" + model + "_part_" + part;
		Writer w;
		if (!zip) {
			path = path + ".txt";
			w = new BufferedWriter(new FileWriter(new File(path), appendOutput));
		} else {
			path = path + ".gz";
			w = new OutputStreamWriter(new GZIPOutputStream(
					new BufferedOutputStream(new FileOutputStream(
							new File(path), appendOutput))), "UTF-8");
		}
		writerMap.put(model, new WriterRecord(w, part, 0));
	}

	protected void writeOutput(String filename, String output) {
		writeOutput(filename, output, true);
	}

	protected void writeOutput(String filename, String output,
			boolean addLineBreak) {
		try {
			if (!writerMap.containsKey(filename)) {
				initFileForModel(filename, 0);
			}
			WriterRecord rec = writerMap.get(filename);
			rec.count += output.getBytes().length;

			if (rec.count > countThresh) {
				closeWriter(rec.writer);
				rec.part = rec.part + 1;
				initFileForModel(filename, rec.part);
			}
			rec = writerMap.get(filename);
			rec.writer.write(output);
			if (addLineBreak) {
				rec.writer.write("\n");
			}
		} catch (Exception e) {
			System.err.println("Error writing Output!");
			e.printStackTrace();
		}

	}

	/**
	 * Get an Annotation class
	 * 
	 * @param aContext
	 *            UIMA Context
	 * @param paramName
	 *            the provided parameter that contains the class String
	 * @param annotationVariable
	 *            Annotation class that is set as parameter
	 */
	@SuppressWarnings("unchecked")
	private void initAnnotationClass(UimaContext aContext, String paramName,
			Class<? extends Annotation> annotationVariable) {

		Object className = aContext.getConfigParameterValue(paramName)
				.toString();
		if (className == null || className.toString().length() == 0) {
			return;
		}
		try {
			annotationVariable = (Class<? extends Annotation>) Class
					.forName(className.toString());
		} catch (Exception e) {
			System.err.println("Annotation class " + className.toString()
					+ "does not exists: " + e.getMessage());
		}
	}

	private class WriterRecord {
		private Writer writer;
		private int count;
		private int part;

		private WriterRecord(Writer writer, int part, int count) {
			this.writer = writer;
			this.part = part;
			this.count = count;
		}
	}

}
