/*******************************************************************************
 * Copyright 2012
 * Technische Universitaet Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.jobimtext.api.webthesaurus;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import net.sf.json.JSONArray;
import net.sf.json.JSONObject;

import org.apache.pig.impl.util.Pair;
import org.jobimtext.api.IThesaurus;
import org.jobimtext.api.struct.Sense;

/**
 * The WebThesaurus is directly accessing a JoBimViz instance for data retrieval
 * @author Eugen Ruppert
 * 
 */
public class WebThesaurus extends WebResource
		implements
		IThesaurus<String, String, List<Pair<String, Double>>, List<Pair<String, Double>>, List<Sense>, List<Sense>, List<Sense>> {

	private static final String DEFAULT_SENSE = "DEFAULT_SENSE";

	public List<Pair<String, List<String>>> getSentenceHoling(String sentence) {
		String response = getAPIResponseByURL("", "s=" + sentence, this
				.getWebConfiguration().getPath() + "ws/holing/");
		return extractJSONResultSentence(response);
	}

	@Override
	public List<Pair<String, Double>> getSimilarTerms(String key) {
		String response = getAPIResponse(key, "jo", "similar");
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getSimilarTerms(String key,
			int numberOfEntries) {
		String response = getAPIResponse(key, "jo", "similar", numberOfEntries);
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getSimilarTerms(String key,
			double threshold) {
		String response = getAPIResponse(key, "jo", "similar", threshold);
		return extractJSONResultList(response);
	}

	@Override
	public Double getSimilarTermScore(String t1, String t2) {
		String response = getAPIResponse(t1, t2, "jo", "similar-score");
		Double score = extractJSONResultScore(response);
		// try to reverse arguments for a non-zero score
		 if (score == 0) {
			 response = getAPIResponse(t2, t1, "jo", "similar-score");
			 score = extractJSONResultScore(response);
		 }
		return score;
	}

	@Override
	public List<Pair<String, Double>> getSimilarContexts(String value) {
		String response = getAPIResponse(value, "bim", "similar");
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getSimilarContexts(String value,
			int numberOfEntries) {
		String response = getAPIResponse(value, "bim", "similar",
				numberOfEntries);
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getSimilarContexts(String value,
			double threshold) {
		String response = getAPIResponse(value, "bim", "similar", threshold);
		return extractJSONResultList(response);
	}

	@Override
	public Long getTermCount(String key) {
		String response = getAPIResponse(key, "jo", "count");
		return extractJSONResultCount(response);
	}

	@Override
	public Long getContextsCount(String key) {
		String response = getAPIResponse(key, "bim", "count");
		return extractJSONResultCount(response);
	}

	@Override
	public Long getTermContextsCount(String key, String val) {
		String[] arg = {key, val};
		String response = getAPIResponse(arg, "jo/bim", "count");
		return extractJSONResultCount(response);
	}

	@Override
	public Double getTermContextsScore(String key, String val) {
		String[] arg = {key, val};
		String response = getAPIResponse(arg, "jo/bim", "score");
		return extractJSONResultScore(response);
	}

	@Override
	public List<Pair<String, Double>> getTermContextsScores(String key) {
		String response = getAPIResponse(key, "jo/bim", "score");
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getTermContextsScores(String key,
			int numberOfEntries) {
		String response = getAPIResponse(key, "jo/bim", "score",
				numberOfEntries);
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getTermContextsScores(String key,
			double threshold) {
		String response = getAPIResponse(key, "jo/bim", "score", threshold);
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getContextsTermScores(String key) {
		String response = getAPIResponse(key, "bim/jo", "score");
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getContextsTermScores(String key,
			int numberOfEntries) {
		String response = getAPIResponse(key, "bim/jo", "score",
				numberOfEntries);
		return extractJSONResultList(response);
	}

	@Override
	public List<Pair<String, Double>> getContextsTermScores(String key,
			double threshold) {
		String response = getAPIResponse(key, "bim/jo", "score", threshold);
		return extractJSONResultList(response);
	}

	@Override
	public List<Sense> getSenses(String key) {
		String response = getAPIResponse(key, "jo", "senses");
		return extractJSONResultSenses(response);
	}

	@Override
	public List<Sense> getIsas(String key) {
		String response = getAPIResponse(key, "jo", "isas");
		return extractJSONResultSenses(response);
	}

	@Override
	public List<Sense> getSenseCUIs(String key) {
		String response = getAPIResponse(key, "jo", "sense-cuis");
		return extractJSONResultSenses(response);
	}

	@Override
	public List<Sense> getSenses(String key, String type) {
		String response = getAPIResponse(key, "jo", "senses", type, true);
		return extractJSONResultSenses(response);
	}

	@Override
	public List<Sense> getIsas(String key, String type) {
		String response = getAPIResponse(key, "jo", "senses", type, true);
		return extractJSONResultSenses(response);
	}

	@Override
	public List<Sense> getSenseCUIs(String key, String type) {
		String response = getAPIResponse(key, "jo", "senses", type, true);
		return extractJSONResultSenses(response);
	}

	@Override
	public String[] getSensesTypes() {
		String response = getAPIResponse("", "jo", "sensestypes");
		return extractJSONResultSenseTypes(response);
	}

	@Override
	public String getStandardSense() {
		String response = getAPIResponse("", "jo", "standardsense");
		return parseJSON(response).toString();
	}

	@Override
	public boolean isTermContained(String key) {
		String response = getAPIResponse(key, "jo", "istermcontained");
		return !JSONArray.fromObject(parseJSON(response)).isEmpty();

	}

	@Override
	public String[] isTermContained(String... key) {
		String response = getAPIResponse(key, "jo", "istermcontained");
		JSONArray resAr = JSONArray.fromObject(parseJSON(response));
		String[] ret = new String[resAr.size()];
		for (int i = 0; i < resAr.size(); i++) {
			ret[i] = resAr.getString(i);
		}
		return ret;
	}

	@Override
	public boolean isContextsContained(String key) {
		String response = getAPIResponse(key, "bim", "iscontextscontained");
		return !JSONArray.fromObject(parseJSON(response)).isEmpty();
	}

	@Override
	public String[] isContextsContained(String... key) {
		String response = getAPIResponse(key, "bim", "iscontextscontained");
		JSONArray resAr = JSONArray.fromObject(parseJSON(response));
		String[] ret = new String[resAr.size()];
		for (int i = 0; i < resAr.size(); i++) {
			ret[i] = resAr.getString(i);
		}
		return ret;
	}

	private List<Pair<String, Double>> extractJSONResultList(String response) {
		List<Pair<String, Double>> res = new ArrayList<Pair<String, Double>>();

		JSONArray results = (JSONArray) parseJSON(response, true);
		if (results.size() < 1) {
			return res;
		}
		
		for (Object j : results) {
			JSONObject js = JSONObject.fromObject(j);
			res.add(new Pair(js.getString("key"), js.getDouble("score")));
		}
		return res;
	}

	private Double extractJSONResultScore(String response) {
		JSONObject res = (JSONObject) parseJSON(response);
		return res.getDouble("score");
	}

	private Long extractJSONResultCount(String response) {
		JSONObject res = (JSONObject) parseJSON(response);
		return res.getLong("count");
	}

	private String[] extractJSONResultSenseTypes(String response) {
		JSONObject res = (JSONObject) parseJSON(response);
		JSONArray resAr = res.getJSONArray("sensetypes");
		String[] ret = new String[resAr.size()];
		for (int i = 0; i < resAr.size(); i++) {
			ret[i] = resAr.getString(i);
		}
		return ret;
	}

	private List<Sense> extractJSONResultSenses(String response) {
		List<Sense> res = new ArrayList<Sense>();

		JSONArray results;
		try {
			results = (JSONArray) parseJSON(response);
		} catch (IllegalStateException e) {			
			throw new IllegalStateException(
					"This model does not support Sense methods!");
		}
		for (Object j : results) {
			Sense sense = new Sense();

			JSONObject js = JSONObject.fromObject(j);
			JSONArray sAr = js.getJSONArray("senses");
			List<String> senses = new ArrayList<String>();
			for (Object s : sAr) {
				senses.add(s.toString());
			}
			sense.setSenses(senses);
			JSONArray isaAr = js.getJSONArray("isas");
			List<String> isas = new ArrayList<String>();
			for (Object i : isaAr) {
				isas.add(i.toString());
			}
			sense.setIsas(isas);
			sense.setCui(js.getString("cui"));
			res.add(sense);
		}
		return res;
	}

	private List<Pair<String, List<String>>> extractJSONResultSentence(
			String response) {
		List<Pair<String, List<String>>> res = new ArrayList<Pair<String, List<String>>>();

		JSONObject parse = JSONObject.fromObject(response);
		JSONArray holings;
		if (parse.get("error").toString().compareTo("null") == 0) {
			holings = parse.getJSONArray("holings");
		} else {
			return null;
		}
		String oldKey = "";
		List<String> bims = new ArrayList<String>();
		String bim = "";
		for (Object j : holings) {
			String key = JSONObject.fromObject(j).getJSONObject("key")
					.getString("key");
			bim = JSONObject.fromObject(j).getString("relationString");
			if (oldKey.compareTo("") != 0 && key.compareTo(oldKey) != 0) {
				res.add(new Pair<String, List<String>>(oldKey, bims));
				bims = new ArrayList<String>();
				bims.add(bim);
			} else {
				bims.add(bim);
			}
			oldKey = key;
		}
		res.add(new Pair<String, List<String>>(oldKey, bims));
		return res;
	}

	private Object parseJSON(String response, boolean list) {
		JSONObject jsonObj;
		try {
			jsonObj = JSONObject.fromObject(response);
		} catch (net.sf.json.JSONException e) {
			throw new IllegalStateException(
					"Invalid JSON result! Method not available for the selected model!");
		}

		if (jsonObj.containsKey("error")
				&& jsonObj.get("error").toString().compareTo("null") == 0) {
			if (list) {
				return jsonObj.get("results");
			} else {
				return jsonObj.get("result");
			}
		}
		return null;
	}

	private Object parseJSON(String response) {
		return parseJSON(response, false);
	}

	private String getAPIResponseByURL(String arguments) {
		return getAPIResponseByURL(arguments, null, this.getWebConfiguration()
				.getAPIPath());
	}

	private String getAPIResponseByURL(String arguments, String parameter) {
		return getAPIResponseByURL(arguments, parameter, this
				.getWebConfiguration().getAPIPath());
	}

	private String getAPIResponseByURL(String arguments, String parameter,
			String apiPath) {
		HttpURLConnection conn;
		BufferedReader rd;
		String line;
		String response = "";
		try {			
			URI uri = new URI(this.getWebConfiguration().getProtocol(), null,
					this.getWebConfiguration().getServer(), this
							.getWebConfiguration().getPort(), apiPath
							+ this.getWebConfiguration().getDataset() + "/"
							+ arguments, parameter, null);
			URL url = new URL(uri.toASCIIString());
			conn = (HttpURLConnection) url.openConnection();
			conn.setRequestMethod("GET");
			rd = new BufferedReader(
					new InputStreamReader(conn.getInputStream(), "UTF-8"));
			while ((line = rd.readLine()) != null) {
				response += line;
			}
			rd.close();
			return response;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	private String getAPIResponse(String key, String type, String method) {
		return getAPIResponseByURL(type + "/" + method + "/" + URLParamEncoder.encode(key));
	}

	private String getAPIResponse(String key, String type, String method,
			int numberOfEntries) {
		return getAPIResponseByURL(type + "/" + method + "/" + URLParamEncoder.encode(key),
				"numberOfEntries=" + numberOfEntries);
	}

	private String getAPIResponse(String key, String type, String method,
			double threshold) {
		return getAPIResponseByURL(type + "/" + method + "/" + URLParamEncoder.encode(key),
				"threshold=" + threshold);
	}

	private String getAPIResponse(String key, String type, String method,
			String senseType, boolean senses) {
		return getAPIResponseByURL(type + "/" + method + "/" + URLParamEncoder.encode(key),
				"sensetype=" + senseType);
	}

	private String getAPIResponse(String[] key, String type, String method) {
		String keys = "";
		for (String k : key) {
			keys += URLParamEncoder.encode(k) + "/";
		}
		return getAPIResponseByURL(type + "/" + method + "/" + keys);
	}

	private String getAPIResponse(String t1, String t2, String type,
			String method) {
		return getAPIResponseByURL(type + "/" + method + "/" + URLParamEncoder.encode(t1)
				+ "/" + URLParamEncoder.encode(t2));
	}

}


class URLParamEncoder {

    public static String encode(String input) {
        StringBuilder resultStr = new StringBuilder();
        for (char ch : input.toCharArray()) {
            if (isUnsafe(ch)) {
                resultStr.append('%');
                resultStr.append(toHex(ch / 16));
                resultStr.append(toHex(ch % 16));
            } else {
                resultStr.append(ch);
            }
        }
        return resultStr.toString();
    }

    private static char toHex(int ch) {
        return (char) (ch < 10 ? '0' + ch : 'A' + ch - 10);
    }

    private static boolean isUnsafe(char ch) {
        if (ch > 128 || ch < 0)
            return true;
        return " %$&+,/:;=?<>#%".indexOf(ch) >= 0;
    }

}