package org.jobimtext.contextualize;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.xml.bind.JAXBException;

import org.apache.uima.UimaContext;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.component.JCasAnnotator_ImplBase;
import org.apache.uima.fit.descriptor.ConfigurationParameter;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.cas.DoubleArray;
import org.apache.uima.jcas.cas.FSArray;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;
import org.jobimtext.api.struct.DatabaseThesaurusDatastructure;
import org.jobimtext.api.struct.IThesaurusDatastructure;
import org.jobimtext.api.struct.Order2;
import org.jobimtext.holing.extractor.JobimAnnotationExtractor;
import org.jobimtext.holing.extractor.JobimExtractorConfiguration;
import org.jobimtext.holing.type.JoBim;
import org.jobimtext.holing.type.Sentence;
import org.jobimtext.type.DistributionalExpansion;
import org.jobimtext.type.DistributionalExpansions;

import com.ibm.bluej.util.common.Pair;

public class SimpleContextualizer extends JCasAnnotator_ImplBase {

	public static final String PARAM_API_CONFIG = "ApiConfig";
	public static final String PARAM_EXTRACTOR_CONFIG = "ExtractorConfig";
	public static final String PARAM_SENTENCE = "SentenceAnnotation";
	public static final String PARAM_NUMBER_OF_EXPANSIONS = "NumberOfExpansions";
	private final int SMOOTH = 1;
	public static final String PARAM_MIN_SIGNIFICANCE = "MinSignificance";


	@ConfigurationParameter(name = PARAM_MIN_SIGNIFICANCE, mandatory = false)
	private double minSignificance = 50.0;
	
	@ConfigurationParameter(name = PARAM_NUMBER_OF_EXPANSIONS)
	private int numberOfEntries;
	@ConfigurationParameter(name = PARAM_API_CONFIG)
	String apiConfig;
	@ConfigurationParameter(name = PARAM_EXTRACTOR_CONFIG)
	String extractorConfig;
	
private JobimAnnotationExtractor extractor;
	@ConfigurationParameter(name = PARAM_SENTENCE, mandatory = false, description = "Specifies the name of the sentence annotation")
	private String sentenceAnnotationClassName = Sentence.class.getName();
	private Class<Annotation> sentenceAnnotationClass;
	private IThesaurusDatastructure<String, String> dt;
	
	@Override
	public void initialize(UimaContext context)
			throws ResourceInitializationException {
		super.initialize(context);
		try {
			extractor = JobimExtractorConfiguration.getExtractorFromXmlFile(extractorConfig);
			dt=  new DatabaseThesaurusDatastructure(apiConfig);
			dt.connect();
			sentenceAnnotationClass = (Class<Annotation>) Class
					.forName(sentenceAnnotationClassName);
			if (sentenceAnnotationClass != null) {
				if (sentenceAnnotationClassName.equals(sentenceAnnotationClass)) {
					sentenceAnnotationClassName = null;
				} else {
					sentenceAnnotationClass = (Class<Annotation>) Class
							.forName(sentenceAnnotationClassName);
				}
			}
		} catch (InstantiationException | IllegalAccessException
				| IllegalArgumentException | InvocationTargetException
				| NoSuchMethodException | SecurityException
				| ClassNotFoundException | JAXBException e) {
			e.printStackTrace();
		}
	}

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		Set<Annotation> filled = new HashSet<Annotation>();
		for (Annotation s : JCasUtil.select(jcas, sentenceAnnotationClass)) {
			for (JoBim jb : JCasUtil.selectCovered(JoBim.class,s)){
				if (!filled.contains(jb.getKey())){
					
					Collection<JoBim> jbs = JCasUtil.selectCovered(JoBim.class, jb.getKey());
					String key = extractor.extractKey(jb);
					List<Order2> exps = dt.getSimilarTerms(key, numberOfEntries);
//					System.out.println(jb.getKey().getCoveredText()+"\t"+jbs.size());
					List<Pair<Order2, Double>> list =  contextualize(jbs,exps);
					if (list.size()==0){
						annotateNoCtx(jb, jcas, exps);
					}else{
						annotateWithContext(jb,jcas,list);
					}

					filled.add(jb.getKey());
				}
			}
		}
	}
	
	private void annotateWithContext(JoBim jb, JCas jcas, List<Pair<Order2, Double>> list) {
		DistributionalExpansions uexps = new DistributionalExpansions(jcas,jb.getKey().getBegin(),jb.getKey().getEnd());
		int i = 0;
		uexps.setEntries(new FSArray(jcas, list.size()));
		for (Pair<Order2, Double> o2 : list){
			DistributionalExpansion e = new DistributionalExpansion(jcas, uexps.getBegin(), uexps.getEnd());
			e.setKey(o2.first.key);
			e.setScore(o2.first.score);
			
			DoubleArray arr = new DoubleArray(jcas,1);
			e.setContextScores(arr);
			arr.set(0, o2.second);
//			e.setContextScore(o2.second);
			uexps.setEntries(i, e);
			i+=1;
		}
		uexps.addToIndexes();
	}

	public void annotateNoCtx(JoBim jb,JCas jcas,List<Order2> exps){
		DistributionalExpansions uexps = new DistributionalExpansions(jcas,jb.getKey().getBegin(),jb.getKey().getEnd());
		int i = 0;
		uexps.setEntries(new FSArray(jcas, exps.size()));
		for (Order2 o2 : exps){
			DistributionalExpansion e = new DistributionalExpansion(jcas, uexps.getBegin(), uexps.getEnd());
			e.setKey(o2.key);
			e.setScore(o2.score);
			DoubleArray arr = new DoubleArray(jcas,1);
			e.setContextScores(arr);
			arr.set(0,0.0);
//			e.setContextScore(0.0);
			uexps.setEntries(i, e);
			i+=1;
		}
		uexps.addToIndexes();
		
	}
	
	public List<Pair<Order2, Double>> contextualize(Collection<JoBim> jobims,List<Order2> exps){
		if (jobims.size() == 0 || jobims.iterator().next() == null) {
			return new ArrayList<Pair<Order2, Double>>();
			
		}

		Annotation key = jobims.iterator().next().getKey();
		
		List<Pair<Order2, Double>> expsScores = new ArrayList<Pair<Order2, Double>>();
		String[] features = new String[jobims.size()];
		@SuppressWarnings("unchecked")
		LinkedList<String>[] featureWordList = new LinkedList[jobims.size()];
		double[] featureWordListNormalized = new double[jobims.size()];
		int i = 0;
		// convert the context to String
		for (JoBim jb : jobims) {
			double norm = 0;
			features[i] = extractor.extractValues(jb);
			featureWordList[i] = new LinkedList<String>();

			for (int j = 0; j < jb.getValues().size(); j++) {
				String fw = extractor.extract(jb.getValues(j));
				featureWordList[i].addLast(fw);
				norm += dt.getTermCount(fw);
			}
			if (norm == 0) {
				// if not in DB, than this word occurs quite rarely
				norm = 100 / 2.0;
			} else {
				norm = 100 / norm / jb.getValues().size();
			}
			featureWordListNormalized[i] = norm;
			i++;

		}
		for (Order2 exp : exps) {// for every lexical expansion

			double z = 1;
			double n = 0;
			for (i = 0; i < features.length; i++) {
				double normalize = featureWordListNormalized[i];
				double score = dt.getTermContextsScore(exp.key, features[i]);
				if (score <= minSignificance) {
					score = 0.0;
				}

//				System.out.println(
//						
//						"Dependency: " + exp.key + "\t" + exp.score + "\t"
//								+ features[i] + "\t" + score + "\t" + normalize
//								+ "\t" + (normalize * (score + 1)) + "\t");
				z *= normalize * (score + SMOOTH);
				n += normalize * (score + SMOOTH);
			}
			if (features.length == 1) {
				expsScores.add(new Pair<Order2, Double>(exp, z));
			} else {
				double score = features.length * z / n + exp.score / 10000.;
				expsScores.add(new Pair<Order2, Double>(exp, score));
			}

//			System.out.println("Pair: " + key.getCoveredText() + "\t" + exp.key + "\t"
//							+ exp.score + "\t" + z + "\t" + n + "\t"
//							+ expsScores.get(expsScores.size() - 1).second);
		}
		if (expsScores.size() == 0) {
			return expsScores;
		}
		Collections.sort(expsScores, new Comparator<Pair<Order2, Double>>() {
			public int compare(Pair<Order2, Double> o1,
					Pair<Order2, Double> o2) {
				return o2.second.compareTo(o1.second);
			}
		});
		return expsScores;
	}

}
