/*******************************************************************************
 * Copyright 2015
 * FG Language Technologie
 * Technische Universitaet Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.jobimtext.hadoop.mapreducer;

//org.jobimtext.hadoop.mapreducer.DruidMweExtractor

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.apache.pig.builtin.MAX;
import org.jobimtext.hadoop.mapreducer.SumReducer.IntSumReducer;

/***
 * 
 * @author Martin Riedl
 * 
 * 
 *         Compute a DRUID Score based on a n-gram Trigram DT with n>1
 *         to run the class you need to specify:
 *         
 *         hadoop jar *jar org.jobimtext.hadoop.mapreducer.DruidMweExtractor dt druid min_druid_score feature_separator normalization_uniqueness normalization_incompleteness min_similarity_score queue
 *         
 *         The method is based on the publication:
 *         Martin Riedl, Chris Biemann (2015): 
 *         A Single Word is not Enough: Ranking Multiword Expressions Using Distributional Semantics. 
 *         In: Proceedings of the 2015 Conference on Empirical Methods in Natural Language Processing (EMNLP 2015). Lisboa, Portugal
 * 
 */
public class DruidMweExtractor {

	@SuppressWarnings("deprecation")
	public static class Map extends MapReduceBase implements
			Mapper<LongWritable, Text, Text, Text> {

		private final IntWritable ONE = new IntWritable(1);
		private boolean useEntryForNormalization = false;
		private double minSimilarity = -1;

		@Override
		public void configure(JobConf job) {
			if (job.get("norm.uq").equals("-1")) {
				useEntryForNormalization = true;
			}

			minSimilarity = Double.parseDouble(job.get("min.sim", "-100.0"));
			super.configure(job);
		}

		public void map(LongWritable key, Text value,
				OutputCollector<Text, Text> output, Reporter reporter)
				throws IOException {

			// split the line separated by tabs
			String[] tokens = value.toString().split("\t");
			String word1 = tokens[0];
			String word2 = tokens[1];
			double score = Double.parseDouble(tokens[2]);
			String feats = tokens[3];

			if (word1.equals(word2)) {
				output.collect(new Text(word1), new Text("0\t0\t" + feats));
			} else {
				if (score < minSimilarity) {
					return;
				}
				if (word2.split(" ").length == 1) {
					output.collect(new Text(word1), new Text("1\t1"));
				} else if (useEntryForNormalization) {
					output.collect(new Text(word1), new Text("0\t1"));
				}
			}

		} // map()

	} // class map

	public static class Reduce extends MapReduceBase implements
			Reducer<Text, Text, Text, Text> {
		private double minScore;
		private double normalizeUq;
		private double normalizeIc;
		private String featureSeparator;

		@Override
		public void configure(JobConf job) {
			
			minScore = Double.parseDouble(job.get("min.score", "-100"));
			normalizeUq = Double.parseDouble(job.get("norm.uq", "-1.0"));
			normalizeIc = Double.parseDouble(job.get("norm.ic", "-1.0"));
			featureSeparator = job.get("feature.separator", "####");

			super.configure(job);
		}

		public void reduce(Text key, Iterator<Text> values,
				OutputCollector<Text, Text> output, Reporter reporter)
				throws IOException {
			double uq = 0.0;
			double normUq = normalizeUq;
			double ic = 0.0;
			double normIc = normalizeIc;
			if (normIc == -1) {
				normIc = 0;
			}
			if (normUq == -1) {
				normUq = 0;
			}

			int maxCount = -1;
			//compute the IC value
			while (values.hasNext()) {
				String v = values.next().toString();
				String[] vs = v.split("\t");
				if (v.trim().length() == 0) {
					continue;
				}

				if (vs.length == 3) {
					// compute IC 
				  // the IC can only be computed for trigram features where the two words are separated by "_@_"
					HashMap<String, Integer> mapCount = new HashMap<String, Integer>();
					// System.out.println(Arrays.toString(vs));
					String[] vss = vs[2].split(featureSeparator);
					normIc = vss.length;
					for (String c : vss) {
						c = c.trim();
						if (c.length() == 0) {
							continue;
						}

						for (String cs : c.split("@")) {
							int count = 0;
							if (mapCount.containsKey(cs)) {
								count = mapCount.get(cs);
							}
							mapCount.put(cs, count + 1);
							if (count + 1 > maxCount) {
								maxCount = count + 1;
							}
						}
					}
				}
				if (normalizeUq == -1) {
					normUq += Integer.parseInt(vs[1]);
				}
				//count the uniqueness for each term
				uq += Integer.parseInt(vs[0]);
			}
			//no negative IC value
			ic = maxCount<=0?0:maxCount;
			if (normalizeIc!=-1){
				normIc = normalizeIc;
			}
			double druid = 1.0 * uq / normUq - 1.0 * ic / normIc;
			if (druid >= minScore) {
				output.collect(new Text(key), new Text("" + druid + "\t" + uq
						+ "\t" + normUq + "\t" + ic + "\t" + normIc));
			}

		}
	}

	@SuppressWarnings("deprecation")
	public static void main(String[] args) throws Exception {
		JobConf conf = HadoopUtil.generateJobConf(args);

		conf.setOutputKeyClass(Text.class);
		conf.setOutputValueClass(Text.class);

		conf.setMapperClass(Map.class);
		// conf.setCombinerClass(Reduce.class);
		conf.setReducerClass(Reduce.class);

		conf.setInputFormat(TextInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);
		if (args.length == 0) {
			System.err
					.println("hadoop jar *jar class input output [feature_separator(####) min_score(-10000) normalize_uq(-1) normalize_ic min_sim(-1)] queue");
			System.err.println("min_score: minimal druid score which is considered");
			System.out.println("normalize_uq: if set to -1 it will normalize the uniqueness by the number of similar terms found for a term. Otherwise it will use the number provided for normalization");
			System.out.println("normalize_ic: if set to -1 it will normalize the incompleteness by the number of significant features for a term. Otherwise it will use the number provided for normalization");
			System.out.println("min_sim:\tminimal smilarity for similar term used to compute the uniqueness");
			System.exit(0);
		}
		FileInputFormat.setInputPaths(conf, new Path(args[0]));
		FileOutputFormat.setOutputPath(conf, new Path(args[1]));

		int maptasks = 120;
		conf.set("min.score", "-10000");
		conf.set("norm.uq", "-1");
		conf.set("norm.ic", "-1");
		conf.set("min.sim", "-1.0");
		conf.set("feature.separator", "####");

		if (args.length > 2)
			conf.set("feature.separator", args[2]);

		if (args.length > 3)
			conf.set("min.score", args[3]);
		if (args.length > 4)
			conf.set("norm.uq", (args[4]));
		if (args.length > 5)
			conf.set("norm.ic", (args[5]));
		if (args.length > 6)
			conf.set("min.sim", (args[6]));

		/* set the maximum number of task per node */
		conf.set("mapred.tasktracker.map.tasks.maximum", "" + maptasks);
		conf.set("mapred.map.tasks", "" + maptasks);
		conf.set("mapred.tasktracker.map", "" + maptasks);

		int reducetasks = 100;

		conf.set("mapred.tasktracker.reduce.tasks.maximum", "" + reducetasks);
		conf.set("mapred.reduce.tasks", "" + reducetasks);
		conf.set("mapred.tasktracker.reduce", "" + reducetasks);

		/*
		 * how much virtual memory the entire process tree of each map/reduce
		 * task will use
		 */
		conf.set("mapred.job.map.memory.mb", "2048");
		conf.set("mapred.job.reduce.memory.mb", "2048");

		conf.set("dfs.replication", "1");
		conf.set("mapred.job.shuffle.input.buffer.percent", "0.2");

		JobClient.runJob(conf);

	} // main

} // class WordCount
