/*******************************************************************************
 * Copyright 2012
 * FG Language Technologie
 * Technische Universitaet Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.jobimtext.hadoop.mapreducer;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.jobimtext.hadoop.mapreducer.SumReducer.IntSumReducer;

/*
 ** Take the context output and count all words
 ** (i.e. the first column multiplied by the number
 ** of occurrences)
 **/
public class RegexUniqMapper {

	@SuppressWarnings("deprecation")
	public static class Map extends MapReduceBase implements
			Mapper<LongWritable, Text, Text, IntWritable> {

		private int[] columns;
		int columnSumup;
		final IntWritable ONE = new IntWritable(1);
		private HashMap<Integer, Pattern> patterns;
		private String regex_separator;

		private HashMap<Integer, Pattern> getPatterns(JobConf job, int[] columns) {
			HashMap<Integer, Pattern> patterns = new HashMap<Integer, Pattern>();
			boolean global = false;
			String regex = "(.*)";
			String reg = "";
			int i = 0;
			while ((reg = job.get("regex.select." + i)) != null) {
				System.out.println(reg);
				String[] regs = reg.split(":");
				if (!regs[0].matches("^[0-9]+$")) {
					global = true;
					regex = regs[1];
					patterns.put(columns[i], Pattern.compile(regex));
					continue;
				}
				int id = Integer.parseInt(regs[0]);

				int idx = Arrays.binarySearch(columns, id);
				if (idx == -1) {
					System.out.println("Column " + id
							+ " for regular expression (" + regs[1]
							+ ") is not used");
					continue;
				}
				patterns.put(id, Pattern.compile(regs[1]));
				System.out.println("Set "+id+" "+regs[1]);
				i += 1;
			}
			System.out.println(patterns);
			if (columns.length != patterns.size()) {
				for (i = 0; i < columns.length; i++) {
					
					if (!patterns.containsKey(columns[i])) {
						System.out.println("Not contained: "+columns[i]);
						if (global) {
							patterns.put(columns[i], Pattern.compile(regex));
							
						} else {
							patterns.put(columns[i], Pattern.compile("(.*)"));
						}
					}
				}
			}
			System.out.println(patterns);
			return patterns;
		}

		@Override
		public void configure(JobConf job) {
			String columns = job.get("columns");
			columnSumup = job.getInt("column.sumup", -1);
			regex_separator = job.get("regex.separator", " ");
			String[] strCols = columns.split(",");
			this.columns = new int[strCols.length];
			int i = 0;
			for (String col : strCols) {
				this.columns[i] = Integer.parseInt(col);
				i++;
			}
			patterns = getPatterns(job, this.columns);

			super.configure(job);
		}

		public void map(LongWritable key, Text value,
				OutputCollector<Text, IntWritable> output, Reporter reporter)
				throws IOException {

			// split the line separated by tabs
			String[] tokens = value.toString().split("\t");
			String newKey = "";
			for (int col : columns) {
				
				Matcher m = patterns.get(col).matcher(tokens[col]);
				String subkey = "";
				if (m.find()) {
					for (int i = 1; i <= m.groupCount(); i++) {
						subkey += regex_separator + m.group(i);
					}
				}
				newKey += subkey.trim() + "\t";
			}
			newKey = newKey.trim();
			if (columnSumup < 0) {
				output.collect(new Text(newKey), ONE);
			} else {
				output.collect(new Text(newKey),
						new IntWritable(Integer.parseInt(tokens[columnSumup])));
			}

		} // map()

	} // class map

	@SuppressWarnings("deprecation")
	public static void main(String[] args) throws Exception {
		System.out.println(Arrays.toString(" ".split(":")));
		Pattern.compile("^([^#]*)#(.*)$");
		if (args.length < 3) {
			System.out.println("Usage: ");
			System.out
					.println("hadoop jar jobimtext.jar org.jobimtext.hadoop.mapreducer.RegexUniqMapper input output columns sumup.column regex_separator [0:]regex [1:regex 2:regex] [queuename]");
			System.out.println("input: input folder");
			System.out.println("output: output folder");
			System.out.println("columns: columns to use e.g. 0,3,5");
			System.out
					.println("sump.column: column with integer, which is summed up");
			System.out
					.println("regex_separator: separator between two matches within a column e.g. '(w.*) (f.*)'");
			System.out
					.println("regex: regular expression selecting which part of each column is used e.g. ^(.*e*)$ will select only words which end with 'e'");
			System.out
					.println("       if specifying a number to the regex (e.g. 1:^(,*e)$ this extraction rule will only apply to that specific column");
			System.out
					.println("       if e.g. columns equals 1,3,5 but only a pattern for 1 is specified, for the remaining fields the complete text is taken");
		}

		JobConf conf = HadoopUtil.generateJobConf(args);

		conf.setOutputKeyClass(Text.class);
		conf.setOutputValueClass(IntWritable.class);

		conf.setMapperClass(Map.class);
		conf.setCombinerClass(IntSumReducer.class);
		conf.setReducerClass(IntSumReducer.class);

		conf.setInputFormat(TextInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);
		FileInputFormat.setInputPaths(conf, new Path(args[0]));
		FileOutputFormat.setOutputPath(conf, new Path(args[1]));

		String columns = args[2];
		String columnSumup = "-1";
		String regex = "(.*)";

		if (args.length > 4) {
			columnSumup = args[3];
		}

		int maptasks = 120;
		conf.set("regex.separator", args[4]);
		conf.set("columns", columns);
		conf.set("column.sumup", columnSumup);
		if (args.length > 6) {
			for (int i = 5; i < args.length - 1; i++) {

				conf.set("regex.select." + (i - 5), args[i]);
				System.out.println("regex.select." + (i - 5) + args[i]);
			}
		}

		/* set the maximum number of task per node */
		conf.set("mapred.tasktracker.map.tasks.maximum", "" + maptasks);
		conf.set("mapred.map.tasks", "" + maptasks);
		conf.set("mapred.tasktracker.map", "" + maptasks);

		int reducetasks = 100;

		conf.set("mapred.tasktracker.reduce.tasks.maximum", "" + reducetasks);
		conf.set("mapred.reduce.tasks", "" + reducetasks);
		conf.set("mapred.tasktracker.reduce", "" + reducetasks);

		/*
		 * how much virtual memory the entire process tree of each map/reduce
		 * task will use
		 */
		conf.set("mapred.job.map.memory.mb", "2048");
		conf.set("mapred.job.reduce.memory.mb", "2048");

		conf.set("dfs.replication", "1");
		conf.set("mapred.job.shuffle.input.buffer.percent", "0.2");

		JobClient.runJob(conf);

	} // main

} // class WordCount
