package org.jobimtext.hadoop.mapreducer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.jobimtext.util.StringUtils;

public class SimCount {

	/**
	 * The reducer step will sum all float values, i.e. the weight for any
	 * (word1,word2) pair sharing a feature.
	 */
	@SuppressWarnings("deprecation")
	public static class Reduce extends MapReduceBase implements
			Reducer<Text, Text, Text, Text> {
		boolean addFeatures = false;
		static final String SEPARATOR="####";
		@Override
		public void configure(JobConf job) {
			super.configure(job);
			addFeatures = job.getBoolean("addFeatures", addFeatures);
		}

		// the whole line coming as value

		/**
		 * Sum up all float values (similarity weights).
		 * 
		 * @param key
		 *            The input key (in this case the (word1,word2) pair).
		 * @param values
		 *            The values coming in (in this step the float weights).
		 * @param output
		 *            The output format: (key, value) same as input
		 * @param reporter
		 *            A facility for Map-Reduce applications to report progress
		 *            and update counters, status information etc.
		 */
		public void reduce(Text key, Iterator<Text> values,
				OutputCollector<Text, Text> output, Reporter reporter)
				throws IOException {

			// reset the sum and the featurelist for each key
			String line;
			double sum = 0.0;

			String features = "";
			// sum all values of this key
			while (values.hasNext()) {
				line = values.next().toString();

				if (addFeatures) {
					int idx = line.indexOf("\t");
					if (line.substring(idx+1).startsWith(SEPARATOR) || features.length()==0){
						features += line.substring(idx + 1);
					}else{
						features += SEPARATOR + line.substring(idx + 1);
					}
					// the weight is the first element
					sum += Double.valueOf(line.substring(0, idx));
				} else {
					sum += Double.valueOf(line);
				}
			} // while

			// begin final result with the weight and then tab
			// then concatenate the feature list
			String endResult = Double.toString(sum);
			if (addFeatures)
				endResult += "\t" + features.trim();

			// output new string: summed weight and expanded list
			output.collect(key, new Text(endResult));

		} // reduce()

	} // class reduce

	@SuppressWarnings("deprecation")
	public static void main(String[] args) throws Exception {
		JobConf conf = HadoopUtil.generateJobConf(args);

		/* set the new defined type to be used */
		conf.setMapOutputKeyClass(Text.class);
		conf.setMapOutputValueClass(Text.class);
		Class<? extends Mapper<?, ?, ?, ?>> mapClass = SimCountMapperOne.class;
		if (args.length > 2) {
			String method = args[2];
			if (method.equals("one")) {
				mapClass = SimCountMapperOne.class;
			} else if (method.equals("scored")) {
				mapClass = SimCountMapperScored.class;
			} else if (method.equals("log_scored")) {
				mapClass = SimCountMapperLogScored.class;
			} else {
				System.err.println("The Mapper methods \"" + method + "\""
						+ " is unknown and the MapOne is used");
			}
		}
		conf.setMapperClass(mapClass);
		conf.setCombinerClass(Reduce.class);
		conf.setReducerClass(Reduce.class);

		conf.setInputFormat(TextInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);

		FileInputFormat.setInputPaths(conf, new Path(args[0]));
		FileOutputFormat.setOutputPath(conf, new Path(args[1]));

		conf.setBoolean("addFeatures", Boolean.parseBoolean(args[3]));
		/* number of milliseconds before killing a not responding task */
		conf.set("mapred.task.timeout", "600000");

		/* change to 128mb */
		conf.set("dfs.block.size", "134217728");

		/* set the maximum number of task per node */
		int maptasks = 200;
		/*
		 * Number of map tasks to deploy on each machine. 0.5 to 2 *
		 * (cores/node)
		 */
		conf.set("mapred.tasktracker.map.tasks.maximum", "" + maptasks);
		conf.set("mapred.tasktracker.map", "" + maptasks);
		/*
		 * The default number of map tasks per job. Typically set to a prime
		 * several times greater than number of available hosts.
		 */
		conf.set("mapred.map.tasks", "" + maptasks);

		int reducetasks = 400;

		conf.set("mapred.tasktracker.reduce.tasks.maximum", "" + reducetasks);
		conf.set("mapred.tasktracker.reduce", "" + reducetasks);
		conf.set("mapred.reduce.tasks", "" + reducetasks);

		/*
		 * how much virtual memory the entire process tree of each map/reduce
		 * task will use
		 */
		conf.set("mapred.job.map.memory.mb", "4000");
		conf.set("mapred.job.reduce.memory.mb", "4000");
		conf.set("mapred.job.shuffle.input.buffer.percent","0.2");
//		conf.set("mapred.map.output.compression.codec", "org.apache.hadoop.io.compress.DefaultCodec");
		conf.set("mapred.reduce.copy.backoff", "300");
		conf.set("io.sort.factor","10");
		conf.set("mapred.job.shuffle.merge.percent","0.66" );
		conf.set("mapred.inmem.merge.threshold", "1000");
		conf.set("mapred.job.reduce.input.buffer.percent", "0.20");
		conf.set("dfs.replication", "1");

		/*
		 * reduce I/O load
		 */
		conf.set("mapred.child.java.opts", "-Xmx1400M");

		conf.set("io.sort.mb", "300");
		conf.set("io.sort.factor", "30");

		JobClient.runJob(conf);

	} // main
}
