/*******************************************************************************
 * Copyright 2012
 * FG Language Technologie
 * Technische Universitaet Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.jobimtext.hadoop.mapreducer;

import java.io.IOException;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.jobimtext.hadoop.mapreducer.SumReducer.LongSumReducer;

/*
 ** Take the context output and count all words
 ** (i.e. the first column multiplied by the number
 ** of occurrences)
 **/
public class UniqMapper {

	public static class Map extends MapReduceBase implements
			Mapper<LongWritable, Text, Text, LongWritable> {

		private int[] columns;
		int columnSumup;
		final LongWritable ONE = new LongWritable(1);

		@Override
		public void configure(JobConf job) {
			String columns = job.get("columns");
			columnSumup = job.getInt("column.sumup", -1);
			String[] strCols = columns.split(",");
			this.columns = new int[strCols.length];
			int i = 0;
			for (String col : strCols) {
				
				this.columns[i] = Integer.parseInt(col);
				i++;
			}

			super.configure(job);
		}

		public void map(LongWritable key, Text value,
				OutputCollector<Text, LongWritable> output, Reporter reporter)
				throws IOException {

			// split the line separated by tabs
			String[] tokens = value.toString().split("\t",-1);
			String newKey = "";
			for (int col : columns) {
				if (tokens.length<=col){
					return;
				}
				newKey += tokens[col] + "\t";
			}
			newKey = newKey.trim();
			if (columnSumup < 0) {
				output.collect(new Text(newKey), ONE);
			}else{
				output.collect(new Text(newKey), new LongWritable(Long.parseLong(tokens[columnSumup])));
			}

		} // map()

	} // class map

	public static void main(String[] args) throws Exception {
		if(args.length<3){
			System.out.println("java ....UniqMapper input_folder output_folder columns_to_uniq_comma_separated [column_to_sumup] queue");
		}
		JobConf conf = HadoopUtil.generateJobConf(args);

		conf.setOutputKeyClass(Text.class);
		conf.setOutputValueClass(LongWritable.class);

		conf.setMapperClass(Map.class);
		conf.setCombinerClass(LongSumReducer.class);
		conf.setReducerClass(LongSumReducer.class);

		conf.setInputFormat(TextInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);

		FileInputFormat.setInputPaths(conf, new Path(args[0]));
		FileOutputFormat.setOutputPath(conf, new Path(args[1]));

		String columns = args[2];
		String columnSumup = "-1";
		if (args.length > 4) {
			columnSumup = args[3];
		}

		int maptasks = 120;

		conf.set("columns", columns);
		conf.set("column.sumup", columnSumup);

		/* set the maximum number of task per node */
		conf.set("mapred.tasktracker.map.tasks.maximum", "" + maptasks);
		conf.set("mapred.map.tasks", "" + maptasks);
		conf.set("mapred.tasktracker.map", "" + maptasks);

		int reducetasks = 100;

		conf.set("mapred.tasktracker.reduce.tasks.maximum", "" + reducetasks);
		conf.set("mapred.reduce.tasks", "" + reducetasks);
		conf.set("mapred.tasktracker.reduce", "" + reducetasks);

		/*
		 * how much virtual memory the entire process tree of each map/reduce
		 * task will use
		 */
		conf.set("mapred.job.map.memory.mb", "2048");
		conf.set("mapred.job.reduce.memory.mb", "2048");

		conf.set("dfs.replication", "1");
		conf.set("mapred.job.shuffle.input.buffer.percent", "0.2");

		JobClient.runJob(conf);

	} // main

} // class WordCount
