package org.jobimtext.holing.annotator;


import org.apache.uima.UimaContext;
import org.apache.uima.analysis_component.JCasAnnotator_ImplBase;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.descriptor.ConfigurationParameter;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.cas.FSArray;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;
import org.apache.uima.util.Logger;
import org.jobimtext.holing.type.Jo;
import org.jobimtext.holing.type.JoBim;
import org.jobimtext.holing.type.Sentence;

/**
 * The BigramHolingAnnotator is used to annotate bigrams in a {@link Sentence}.
 * 
 */
public class BigramHolingAnnotator extends JCasAnnotator_ImplBase {
	public static final String PARAM_SENTENCE_ANNOTATION = "SentenceAnnotation";
	public static final String PARAM_TOKEN_ANNOTATION = "TokenAnnotation";
	public static final String PARAM_JO_ANNOTATION = "JoAnnotation";
	public static final String PARAM_BIM_ANNOTATION = "BimAnnotation";
	public static final String PARAM_ALLOW_EMPTY_BIMS_FOR_JO = "allowEmptyBimsForJo";

	private Logger logger;

	@ConfigurationParameter(name = PARAM_ALLOW_EMPTY_BIMS_FOR_JO, mandatory = false)
	private boolean allowEmptyBimsForJo = false;

	@ConfigurationParameter(name = PARAM_SENTENCE_ANNOTATION)
	private Class<? extends Annotation> sentenceAnnotation = org.jobimtext.holing.type.Sentence.class;
	@ConfigurationParameter(name = PARAM_TOKEN_ANNOTATION)
	private Class<? extends Annotation> tokenAnnotation = org.jobimtext.holing.type.Token.class;
	@ConfigurationParameter(name = PARAM_JO_ANNOTATION)
	private Class<? extends Annotation> joAnnotation = Jo.class;
	@ConfigurationParameter(name = PARAM_BIM_ANNOTATION)
	private Class<? extends Annotation> bimAnnotation = org.jobimtext.holing.type.Token.class;

	/**
	 * Initialize the annotator.
	 * 
	 * @param aContext
	 *            UIMA Annotator context
	 */
	@Override
	public void initialize(UimaContext aContext)
			throws ResourceInitializationException {
		super.initialize(aContext);
		logger = aContext.getLogger();
		logger.log(Level.FINE, this.getClass().getName() + ":initialize");
		
		sentenceAnnotation = initAnnotationClass(aContext,
				PARAM_SENTENCE_ANNOTATION, sentenceAnnotation);
		tokenAnnotation = initAnnotationClass(aContext, PARAM_TOKEN_ANNOTATION,
				tokenAnnotation);
		joAnnotation = initAnnotationClass(aContext, PARAM_JO_ANNOTATION,
				joAnnotation);
		bimAnnotation = initAnnotationClass(aContext, PARAM_BIM_ANNOTATION,
				bimAnnotation);
	}

	@SuppressWarnings("unchecked")
	private Class<? extends Annotation> initAnnotationClass(
			UimaContext aContext, String paramName,
			Class<? extends Annotation> annotationVariable) {

		Object className = aContext.getConfigParameterValue(paramName)
				.toString();
		if (className == null || className.toString().length() == 0) {
			return null;
		}
		try {
			annotationVariable = (Class<? extends Annotation>) Class
					.forName(className.toString());

		} catch (Exception e) {
			System.err.println("Annotation class " + className.toString()
					+ "does not exists: " + e.getMessage());
		}
		return annotationVariable;
	}

	/**
	 * Process all sentences in the CAS, annotate bigrams.
	 * 
	 * @param aJCas
	 *            JCas that should be processed
	 */
	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		logger.log(Level.INFO, "BigramHolingAnnotator: Start");

		for (Annotation s : JCasUtil.select(jcas, sentenceAnnotation)) {
			Annotation leftNeighbor = null;
			for (Annotation t : JCasUtil.selectCovered(tokenAnnotation, s)) {
				// process only single-word Tokens, skip annotated Jos
				if (t instanceof Jo) {
					continue;
				}

				// preceding Token must be present for bigrams
				if (leftNeighbor == null) {
					leftNeighbor = t;
					continue;
				}

				addBigram(leftNeighbor, t, jcas);
				leftNeighbor = t;

			}

		}
		logger.log(Level.INFO, "BigramHolingAnnotator: Stop");
	}

	public void addBigram(Annotation left, Annotation right, JCas jcas) {
		// bigram for right token
		JoBim jb = new JoBim(jcas, right.getBegin(), right.getEnd());
		jb.setRelation("-Bigram");
		jb.setKey(right);
		jb.setHole(1);

		FSArray arr = new FSArray(jcas, 1);
		arr.set(0, left);
		jb.setValues(arr);
		jb.addToIndexes();

		// bigram for left token
		jb = new JoBim(jcas, left.getBegin(), left.getEnd());
		jb.setRelation("Bigram");
		jb.setKey(left);
		jb.setHole(0);

		arr = new FSArray(jcas, 1);
		arr.set(0, right);
		jb.setValues(arr);
		jb.addToIndexes();
	}

}
