/*******************************************************************************
* Copyright 2012
* Copyright (c) 2012 IBM Corp.
* 
* and
* 
* FG Language Technology
* Technische Universitaet Darmstadt
* 
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*   http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
******************************************************************************/
package org.jobimtext.holing.annotator;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.uima.UimaContext;
import org.apache.uima.analysis_component.JCasAnnotator_ImplBase;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.cas.impl.FeatureValuePathImpl;
import org.apache.uima.fit.descriptor.ConfigurationParameter;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;
import org.apache.uima.util.Logger;
import org.jobimtext.holing.type.Jo;
import org.jobimtext.holing.type.Sentence;
import org.jobimtext.holing.type.Token;

/**
 * The JoAnnotator is used to annotate n-grams up to the size '{@link #size}' in a {@link Sentence}.
 *
 */
public class JoAnnotator extends JCasAnnotator_ImplBase {

	public static final String PARAM_JO_TOKEN_LENGTH = "joTokenLength";
	public static final String PARAM_SENTENCE_ANNOTATION="SentenceAnnotation";
	public static final String PARAM_TOKEN_ANNOTATION="TokenAnnotation";
	
	@ConfigurationParameter(name = PARAM_SENTENCE_ANNOTATION)
	protected Class<? extends Annotation> sentenceAnnotation = Sentence.class;
    @ConfigurationParameter(name = PARAM_TOKEN_ANNOTATION)
    protected Class<? extends Annotation> tokenAnnotation = Token.class;
    
	private Logger logger;
	
	/**
	 * The maximal number of Tokens for each Jo, default 1. Used to specify n-gram sizes.
	 */
	private int size = 1;

	/**
	 * Initialize the annotator.
	 * 
	 * @param aContext UIMA Annotator context
	 */
	@Override
	public void initialize(UimaContext context)
			throws ResourceInitializationException {
		super.initialize(context);
		logger = context.getLogger();
		logger.log(Level.FINE, this.getClass().getName() + ":initialize");
		size = (Integer) context.getConfigParameterValue(PARAM_JO_TOKEN_LENGTH);
		sentenceAnnotation=initAnnotationClass(context, PARAM_SENTENCE_ANNOTATION , sentenceAnnotation);
		tokenAnnotation=initAnnotationClass(context, PARAM_TOKEN_ANNOTATION , tokenAnnotation);
		
	}
	@SuppressWarnings("unchecked")
	private Class<? extends Annotation> initAnnotationClass(
			UimaContext aContext, String paramName,Class<? extends Annotation> annotationVariable) {
		
		Object className  = aContext.getConfigParameterValue(paramName);
		if(className==null ||className.toString().length()==0){
			return null; 
		}
		try {
			annotationVariable = (Class<? extends Annotation>) Class
					.forName(className.toString());
			
		} catch (Exception e) {
			System.err.println("Annotation class "+className.toString()+"does not exists: "+e.getMessage());
		}
		return annotationVariable;
	}


	/**
	 * Process all sentences in the CAS, annotate n-gram Jos.
	 * 
	 * @param aJCas JCas that should be processed
	 */
	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		logger.log(Level.INFO, "JoAnnotator: Start");
		for (Annotation s : JCasUtil.select(jcas, sentenceAnnotation)) {
			LinkedList<Annotation> list = new LinkedList<Annotation>();
			Set<String> set = new HashSet<String>();
			
			// process all Tokens in the Sentence,
			// add new Tokens to the list and generate n-gram Jos from list
			for (Annotation t : JCasUtil.selectCovered(tokenAnnotation, s)) {				
				// skip Jo tokens
//				System.out.println("T\t"+t.getCoveredText());
				if (t instanceof Jo)
					continue;
				addToken(list,t);
				//list.add(t);
				addList(list,set,jcas);
			}
		}
		logger.log(Level.INFO, "JoAnnotator: Stop");
	}
	
	public List<Annotation> addToken(LinkedList<Annotation> list, Annotation t){
		list.add(t);
		return list;
	}

	/**
	 * Create Jo Annotations from a Token list. 
	 * All possible n-gram Jos up to {@link #size} are Annotated.
	 * 
	 * @param list list of Tokens
	 * @param set set of index positions of already added Jos
	 * @param jcas target CAS
	 */
	private void addList(LinkedList<Annotation> list, Set<String> set,JCas jcas) {
		// choose all possible n-gram sizes from 1 to 'size' to annotate n-gram Jos
		for(int i=1;i<=Math.min(size,list.size());i++){
			// create index positions for the last n tokens
			int begin =list.get(list.size()-i).getBegin();
			int end =list.get(list.size()-1).getEnd();
			
			// if Tokens already marked as annotated, return
			String idx = begin + ":" + end;
			if (set.contains(idx))
				return;
			
			// create new n-gram Jo, mark index positions as annotated			
			Jo jo = new Jo(jcas);
			jo.setBegin(begin);
			jo.setEnd(end);
			jo.addToIndexes();
//			System.out.println(i+"\t"+jo.getCoveredText());
			set.add(idx);
			//System.out.println("JO"+ jo.getCoveredText());
			
		}
		// if the list contains the maximum number of tokens, 
		// remove first element to save memory
		if(list.size()==size){
			list.removeFirst();
		}
	}

//	public void process(JCas jcas, final int length) {
//		LinkedList<Annotation> al = new LinkedList<Annotation>();
//		for (Sentence s : select(jcas, Sentence.class)) {
//			Set<String> set = new HashSet<String>();
//			for (Token t : selectCovered(Token.class, s)) {
//				if (t instanceof Jo)
//					continue;
//				al.add(t);
//				addJo(jcas, al, length, set);
//			}
//		}
//	}
//
//	public void addJo(JCas jcas, LinkedList<Annotation> al, int length,
//			Set<String> sets) {
//		if (al.size() != length) {
//			return;
//		}
//		String idx = al.getFirst().getBegin() + ":" + al.getLast().getEnd();
//		if (sets.contains(idx))
//			return;
//		sets.add(idx);
//		Jo jo = new Jo(jcas);
//		jo.setBegin(al.getFirst().getBegin());
//		jo.setEnd(al.getLast().getEnd());
//		jo.addToIndexes();
//		al.removeFirst();
//	}

}
