/*******************************************************************************
* Copyright 2012
* Copyright (c) 2012 IBM Corp.
* 

* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*   http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
******************************************************************************/
package org.jobimtext.holing.output;

import static org.apache.uima.fit.util.JCasUtil.select;

import java.beans.PropertyDescriptor;
import java.io.File;
import java.util.HashSet;

import org.apache.uima.UimaContext;
import org.apache.uima.analysis_component.JCasAnnotator_ImplBase;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.descriptor.ConfigurationParameter;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;
import org.apache.uima.util.Logger;
import org.jobimtext.holing.extractor.JobimAnnotationExtractor;
import org.jobimtext.holing.extractor.JobimExtractorConfiguration;
import org.jobimtext.holing.type.JoBim;



public abstract class HolingOutputIndexed extends JCasAnnotator_ImplBase {

	public static final String PARAM_EXTRACTOR_CONFIGURATION_FILE = "ExtractorConfigurationFile";
	public static final String PARAM_OUTPUT_PATH = "OutputPath";
	public static final String PARAM_INPUT_NAME = "InputName";
	public static final String PARAM_OUTPUT_SUFFIX = "OutputSuffix";
	public static final String PARAM_DOCUMENT_ANNOTATION = "DocumentAnnotation";
	public static final String PARAM_DOCUMENT_ANNOTATION_ID = "DocumentAnnotationId";
	public static final String PARAM_OVERWRITE_OUTPUT = "overwriteOutput";
	
	public static final String PARAM_FILE_ENCODING = "FileEncoding";
	@ConfigurationParameter(name = PARAM_FILE_ENCODING, defaultValue="UTF-8")
	protected String fileEncoding = "UTF-8";

	protected String outputPath;
	protected String inputName;
	protected String outputSuffix;
	protected String extractorConfFileName;
	protected JobimAnnotationExtractor extractor;
	protected Logger logger;
	protected String documentAnnotationId;
	protected Class<Annotation> documentAnnotationClass;
	protected boolean overwriteOutput = true;
	
	@Override
	public void initialize(UimaContext context)
			throws ResourceInitializationException {
		super.initialize(context);
		logger = context.getLogger();
		logger.log(Level.FINE, "FeatureCounter:initialize");
		inputName = (String) context
				.getConfigParameterValue(PARAM_INPUT_NAME);
		outputPath = (String) context
				.getConfigParameterValue(PARAM_OUTPUT_PATH);
		outputSuffix = (String) context
				.getConfigParameterValue(PARAM_OUTPUT_SUFFIX);
		extractorConfFileName = (String) context
				.getConfigParameterValue(PARAM_EXTRACTOR_CONFIGURATION_FILE);
		String documentAnnotation = (String) context
				.getConfigParameterValue(PARAM_DOCUMENT_ANNOTATION);
		documentAnnotationId = (String) context
				.getConfigParameterValue(PARAM_DOCUMENT_ANNOTATION_ID);
		if (context.getConfigParameterValue(PARAM_OVERWRITE_OUTPUT) != null) {
			overwriteOutput = (boolean) context.getConfigParameterValue(PARAM_OVERWRITE_OUTPUT);
		}
		if (context.getConfigParameterValue(PARAM_FILE_ENCODING) != null) {
			fileEncoding = (String) context.getConfigParameterValue(PARAM_FILE_ENCODING);
		}
		

		try {
			System.out.println("Output Extractor configuration: " + extractorConfFileName);
			extractor = JobimExtractorConfiguration
					.getExtractorFromXmlFile(extractorConfFileName);
		} catch (Exception e) {
			throw new ResourceInitializationException(e);
		}
		try {
			this.documentAnnotationClass = (Class<Annotation>) Class
					.forName(documentAnnotation);
		} catch (Exception e) {
			
			System.err.println("Document Annotation does not exists: "+e.getMessage());
		}
	}

	protected String getvalue(Annotation instance) {
		Object value;
		try {
			value = new PropertyDescriptor(documentAnnotationId,
					documentAnnotationClass).getReadMethod().invoke(instance);
			return value.toString();
		} catch (Exception e) {
			System.err.println("The attribute: " + documentAnnotationId
					+ " could not be found in object of class "
					+ documentAnnotationClass);
			e.printStackTrace();
		}
		return "_Attribute_not_found_";
	}

	@Override
	public void process(JCas aJCas) throws AnalysisEngineProcessException {
		synchronized (this) {
			logger.log(Level.INFO, "Start HolingOutputWriter");
			
			// set output file name
			String unitId="";
			if (inputName != null) {
				// provided name via parameter
				unitId = new File(inputName).getName();
			} else if (documentAnnotationClass != null) {
				Annotation doc = JCasUtil.selectSingle(aJCas,
						documentAnnotationClass);		
				// file name from document Annotation
				unitId = getvalue(doc);
			} else {
				logger.log(Level.SEVERE, "Neither DocumentAnnotation parameter, nor InputName parameter specified");
				throw new AnalysisEngineProcessException();
			}
			
			// replace characters that can be problematic in file names
			unitId = unitId.replace('\\', '-');
			unitId = unitId.replace('/', '-');
//			unitId = unitId.replace('.', '-');
			unitId = unitId.replace('\t', '-');
			
			if (!unitId.matches("[a-zA-Z0-9_-]*")) {
				logger.log(Level.WARNING, "unitId derived from input file name contains special characters! This might cause problems.");
			}
			
	
			HashSet<String> filesToClose = new HashSet<String>();
			for (JoBim jobim : select(aJCas, JoBim.class)) {
				String filename = unitId;
				if (outputSuffix != null) {
					filename += "_"+outputSuffix;
				}
				writeJoBimString(filename, unitId, jobim);
				filesToClose.add(filename);
			}			
			for (String filename : filesToClose) {
			  closeWriter(filename);
			}
			logger.log(Level.INFO, "Finish HolingOutputWriter");	
		}
		

	}

  protected abstract void writeJoBimString(String filename, String unitId, JoBim jobim);
  
  protected abstract void closeWriter(String filename);

}
